import pandas as pd
import numpy as np
import torch
from copy import copy

import matplotlib.pyplot as plt
import seaborn as sns

from .constants import mmoll_mgdl

sns.set()
sns.set_context('paper')
sns.set_style('white')

def clarke_error_grid(output, target, unit:str='mmoll'):
    """
    Return masks for points belonging to certain areas in the Clarke Error Grid 

    Zone A: Clinically Accurate
        Zone A represents glucose values that deviate from the reference by no more than 20% 
        or are in the hypoglycemic range (<70 mg/dl) when the reference is also <70 mg/dl.

    Zone B: Clinically Acceptable
        Upper and lower zone B represents values that deviate from the reference by >20%.

    Zone C: Overcorrecting
        Zone C values would result in overcorrecting acceptable blood glucose levels; 
        such treatment might cause the actual blood glucose to fall below 70 mg/dl or rise above 180 mg/dl.

    Zone D: Failure to Detect
        Zone D represents "dangerous failure to detect and treat" errors. Actual glucose values are 
        outside of the target range, but patient-generated values are within the target range.

    Zone E: Erroneous treatment
        Zone E is an "erroneous treatment" zone. Patient-generated values within this zone are opposite 
        to the reference values, and corresponding treatment decisions would therefore be opposite to that called for.


    Reference:
    [1] Clarke, W.L., Cox, D., Gonder-Frederick, L. A., Carter, W., & Pohl, S. L. (1987). 
        "Evaluating Clinical Accuracy of Systems for Self-Monitoring of Blood Glucose".
        Diabetes Care, 10(5), 622–628. doi: 10.2337/diacare.10.5.622
    [2] Clarke, WL. (2005). "The Original Clarke Error Grid Analysis (EGA)."
        Diabetes Technology and Therapeutics 7(5), pp. 776-779.
    """
    assert unit == 'mmoll' or unit == 'mgdl', 'Please give either mmoll or mgdl as unit here.'
    
    if unit == 'mmoll':
        target = copy(target) * mmoll_mgdl
        output = copy(output) * mmoll_mgdl

    clarke = {}
    clarke['A'] = (abs(target - output) <= 0.2*target) | ((target < 70) & (output < 70))

    clarke['E'] = ((target >= 180) & (output <= 70)) | ((output >= 180) & (target <= 70)) \
        & ~clarke['A']

    clarke['D'] = ((output >= 70) & (output <= 180) & ((target > 240) | (target < 70)))\
        & ~clarke['A'] & ~clarke['E']

    clarke['C'] = ((target >= 70) & (output >= target + 110)) | ((target <= 180) & (output <= (7/5)*target-182))\
        & ~clarke['A'] & ~clarke['E'] & ~clarke['D']

    clarke['B'] = (abs(target - output) > 0.2*target)\
        & ~clarke['A'] & ~clarke['E'] & ~clarke['D'] & ~clarke['C']

    return dict(sorted(clarke.items()))

def format_clarke_error_grid(ax, lims=[0,450], unit='mmoll'):
    lines = np.array([[lims, lims],
        [[lims[0], 70/1.2],     [70, 70]],
        [[70, 70],              [lims[0], 70*0.8]],
        [[70, lims[1]],         [70*0.8, lims[1]*0.8]],
        [[70/1.2, lims[1]/1.2], [70, lims[1]]],
        [[lims[0], 70],         [180, 180]],
        [[180, 180],            [lims[0], 70]],
        [[180, lims[1]],        [70, 70]],
        [[70, 70],              [70*1.2, lims[1]]],
        [[240, 240],            [70, 180]],
        [[240, lims[1]],        [180, 180]],
        [[182*(5/7), 180],      [lims[0], 70]],
        [[70, lims[1]],         [180, lims[1]+110]]])

    s = 1/mmoll_mgdl if unit == 'mmoll' else 1

    lines *= s

    for x, y in lines:
        ax.plot(x, y, c='k')

    ax.set_xlim(np.array(lims)*s)
    ax.set_ylim(np.array(lims)*s)

    texts = np.array([(30, 15, 'A'),
        (414, 260, 'B'),
        (324, 414, 'B'),
        (160, 414, 'C'),
        (160, 15, 'C'),
        (30, 140, 'D'),
        (414, 120, 'D'),
        (30, 260, 'E'),
        (414, 15, 'E'),
        ])
    texts = pd.DataFrame(texts, columns=['x', 'y', 'text'])
    texts[['x', 'y']] = texts[['x', 'y']].astype(float) * s

    for _, (x, y, text) in texts.iterrows():
        ax.text(x, y, text, fontsize=15)

def plot_clarke_error_grid(output, target, unit='mmoll', save_to=None):
    clarke = clarke_error_grid(output, target, unit)
    clarke = pd.DataFrame({c: mask.flatten().numpy() for c, mask in clarke.items()})

    colors = sns.color_palette('RdYlGn_r', len(clarke.keys()))
    colors[2] = (246/256, 245/256, 128/256)#(1, 1, 128/256)
    colors = {k:c for k, c in zip(clarke.keys(), colors)}

    df = pd.DataFrame(torch.stack([output.flatten(), target.flatten()]).T.numpy(), columns=['output', 'target'])
    df['clarke'] = (clarke == True).idxmax(axis=1)

    fig, ax = plt.subplots(figsize=(8,7))
    # note: we cannot use scatterplot in one line because it does not use hue correctly when areas of the CEG are missing
    for k in clarke.keys():
        sns.scatterplot(data=df[df['clarke'] == k], x='target', y='output', color=colors[k], alpha=.5, s=5)
    format_clarke_error_grid(ax, unit=unit)

    plt.xlabel('True Glucose Concentration (mmol/L)')
    plt.ylabel('Predicted Glucose Concentration (mmol/L)')

    #plt.legend(loc='upper center', ncol=5, bbox_to_anchor=(0.5, 1.06), fancybox=False)
    plt.title(f"Clarke Error Grid\n"+'    '.join(["{:s}: {:.2f}%".format(k, clarke[k].sum()/len(df)*100) for k in clarke.keys()]), )

    if save_to:
        plt.savefig(save_to+'.pdf', bbox_inches='tight')
        plt.savefig(save_to+'.png', bbox_inches='tight', dpi=300)
    plt.show()
    plt.close()


import pandas as pd
import numpy as np

from .constants import glucose_levels, mmoll_mgdl

def minimum_eventtime(X:pd.Series,  mintime:str):
    """
    Set event to false if minimum event time is violated
    Note: a pandas rolling function does not work here, as it would not take into account the entire event

    X       - pd.Series with dtype int indicating event (1 or 0)
    mintime - minimum even time
    """
    for start, end in zip(X[X.diff() == 1].index, X[X.diff().shift(-1) == -1].index):        
        assert start <= end, f"Oops, there is something wrong with your event list. Start {start} is larger than end {end}."
        if end - start < pd.to_timedelta(mintime):
            X.loc[(X.index >= start) & (X.index <= end)] = 0
    return X

def hypo(X:pd.Series, mintime:str='14min45S', unit:str='mgdl'):
    """
    Calculate hypo according to definition of https://doi.org/10.2337/dc17-1600
    Note: make sure your data has timestamp indices
    """
    if unit == 'mmoll':
        res = (X * mmoll_mgdl < glucose_levels['target'][0]).astype(int)
    elif unit == 'mgdl':
        res = (X < glucose_levels['target'][0]).astype(int)
    res = minimum_eventtime(res, mintime)
    res.loc[X.isna()] = np.nan
    return res

def hyper(X:pd.Series, mintime:str='14min45S', unit:str='mgdl'):
    """
    Calculate hyper according to definition of https://doi.org/10.2337/dc17-1600
    Note: make sure your data has timestamp indices
    """
    if unit == 'mmoll':
        res = (X * mmoll_mgdl > glucose_levels['target'][1]).astype(int)
    elif unit == 'mgdl':
        res = (X > glucose_levels['target'][1]).astype(int)
    res = minimum_eventtime(res, mintime)
    res.loc[X.isna()] = np.nan
    return res

def symmetric_scale(X:pd.Series, unit:str='mgdl'):
    # symmetric scaling for blood glucose
    if unit == 'mgdl':
        return 1.509*(np.log(X)**1.084 - 5.381)
    elif unit == 'mmoll':
        return 1.794*(np.log(X)**1.026 - 1.861)

def LBGI(X:pd.Series):
    # https://doi.org/10.2337/diacare.21.11.1870
    return symmetric_scale(X).apply(lambda x: 10*x**2 if x < 0 else 0).mean()

def HBGI(X:pd.Series):
    # https://doi.org/10.2337/dc06-1085
    return symmetric_scale(X).apply(lambda x: 10*x**2 if x >= 0 else 0).mean()

def time_in_level(X:pd.Series, l:str, levels:dict=glucose_levels):
    """
    Return number of measurements within glucose range l
    """
    return ((X >= levels[l][0]) & (X <= levels[l][1])).sum()

def perc_in_level(X:pd.Series, l:str, levels:dict=glucose_levels):
    """
    Return percentage of measurements within glucose range l
    """
    return ((X >= levels[l][0]) & (X <= levels[l][1])).sum() / X.count() * 100

def stats_cgm(X:pd.DataFrame, col:str='Glucose Value (mg/dL)'):
    return {'time_in_hypo'     : time_in_level(X[col], 'hypo'),
            'time_in_hypoL2'   : time_in_level(X[col], 'hypo L2'),
            'time_in_hypoL1'   : time_in_level(X[col], 'hypo L1'),
            'time_in_target'   : time_in_level(X[col], 'target'),
            'time_in_hyper'    : time_in_level(X[col], 'hyper'),
            'time_in_hyperL1'  : time_in_level(X[col], 'hyper L1'),
            'time_in_hyperL2'  : time_in_level(X[col], 'hyper L2'),
            'perc_in_hypo'     : perc_in_level(X[col], 'hypo'),
            'perc_in_hypoL2'   : perc_in_level(X[col], 'hypo L2'),
            'perc_in_hypoL1'   : perc_in_level(X[col], 'hypo L1'),
            'perc_in_target'   : perc_in_level(X[col], 'target'),
            'perc_in_hyper'    : perc_in_level(X[col], 'hyper'),
            'perc_in_hyperL1'  : perc_in_level(X[col], 'hyper L1'),
            'perc_in_hyperL2'  : perc_in_level(X[col], 'hyper L2'),
            'glucose_mean'     : X[col].mean(),
            'glucose_std'      : X[col].std(),
            'glucose_cv'       : X[col].std() / X[col].mean() * 100,
            'glucose_rate'     : X['glucose_rate'].mean(),
            'completeness'     : X[col].count() / X['timestamp'].count(),
            'count'            : X[col].count(),
            'LBGI'             : LBGI(X[col]),
            'HBGI'             : HBGI(X[col]),
            'AUC'              : np.trapz(y=X[col], x=X['timestamp']) / np.timedelta64(5, 'm'),
            'hypo'             : X['hypo'].any(),
            'hyper'            : X['hyper'].any()}


import pandas as pd
from cyclingstats.stats import calc_hr_zones, calc_power_zones, agg_zones, agg_power 
from cyclingstats.stats import chronic_training_load, acute_training_load, training_stress_balance

# read time series of power and/or heart rate
df = pd.read_csv("PATH_TO_YOUR_HEARTRATE_AND_POWER_DATA")
df['date'] = pd.to_datetime(df['timestamp'].dt.date)
# perform any other preprocessing steps here

# ---------- zones
# define LTHR and FTP to calculate custom Coggan heart rate and power zones
LTHR = # TODO: fill in a number for the lactate threshold heart rate [bpm]
FTP = # TODO: fill in a number for the functional threshold power [W]

hr_zones = calc_hr_zones(LTHR)
power_zones = calc_power_zones(FTP)

# calculate hr and power zones
df_zones = df.groupby('date').apply(agg_zones, hr_zones=hr_zones, power_zones=power_zones)

# ---------- power
df = df.set_index('timestamp')

# calculate power statistics
df_power = df.groupby('date').apply(agg_power, FTP=FTP)

# fill up dates for which we don't have an entry to get exponential weighted mean (ewm)
dates = df_power.index
df_power = df_power.reindex(date_range)

# calculate ctl, atl and tsb
df_power['chronic_training_load'] = chronic_training_load(df_power['training_stress_score'])
df_power['acute_training_load'] = acute_training_load(df_power['training_stress_score'])
df_power['training_stress_balance'] = training_stress_balance(df_power['chronic_training_load'], df_power['acute_training_load'])

# get back to indices for which there is a training session
df_power = df_power.loc[dates]


from setuptools import setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='glucose-stats',
    version='0.0.7',
    author='Eva van Weenen',
    author_email='evanweenen@ethz.ch',
    description='Calculate glucose statistics from continuous glucose monitoring data',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/evavanweenen/glucose-stats',
    license='MIT',
    packages=['glucosestats'],
    scripts=['bin/run.py'],
    zip_safe=False,
    install_requires=['numpy', 'pandas','torch','matplotlib','seaborn']
    )